# coding: utf-8
"""

"""
import csv
import datetime
import decimal
import functools
import io
import re
import typing

import babel
from flask_babel import _
from babel.numbers import parse_decimal
from flask_login import current_user
import pint
import pytz

from ...logic import schemas, languages
from ...logic.units import ureg, int_ureg
from ...logic.errors import ValidationError


class FormParserT(typing.Protocol):
    def __call__(
            self,
            form_data: typing.Dict[str, typing.List[str]],
            schema: typing.Dict[str, typing.Any],
            id_prefix: str,
            errors: typing.Dict[str, str],
            *,
            file_names_by_id: typing.Dict[int, str],
            required: bool = False,
            previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
    ) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
        ...


def form_data_parser(func: FormParserT) -> FormParserT:
    @functools.wraps(func)
    def wrapper(
            form_data: typing.Dict[str, typing.List[str]],
            schema: typing.Dict[str, typing.Any],
            id_prefix: str,
            errors: typing.Dict[str, str],
            *,
            file_names_by_id: typing.Dict[int, str],
            required: bool = False,
            previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
    ) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
        try:
            return func(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
        except ValueError as e:
            for name in form_data:
                if name.startswith(id_prefix) and '__' not in name[len(id_prefix) + 1:]:
                    errors[name] = str(e)
            return None
        except ValidationError as e:
            for name in form_data:
                if name.startswith(id_prefix) and '__' not in name[len(id_prefix) + 1:]:
                    errors[name] = e.message
            return None
    return wrapper


@form_data_parser
def parse_any_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
    if schema.get('type') == 'object':
        return parse_object_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'array':
        return parse_array_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'text':
        return parse_text_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'sample':
        return parse_sample_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'measurement':
        return parse_measurement_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'object_reference':
        return parse_object_reference_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'datetime':
        return parse_datetime_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'bool':
        return parse_boolean_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'quantity':
        return parse_quantity_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'tags':
        return parse_tags_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'hazards':
        return parse_hazards_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'user':
        return parse_user_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'plotly_chart':
        return parse_plotly_chart_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'timeseries':
        return parse_timeseries_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    elif schema.get('type') == 'file':
        return parse_file_form_data(form_data, schema, id_prefix, errors, required=required, file_names_by_id=file_names_by_id, previous_data=previous_data)
    raise ValueError('invalid schema')


@form_data_parser
def parse_text_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys == [id_prefix + '__text']:
        text_list = form_data.get(id_prefix + '__text', [])
        text = text_list[0] if text_list else None
        if not text and not required:
            return None
        # an empty text should get a min length error from validation, but a
        # missing text can be omitted if there is a non-zero min length
        if text is None and schema.get('minLength', 0) > 0:
            return None
        if text is None:
            text = ""
        data: typing.Dict[str, typing.Any] = {
            '_type': 'text',
            'text': str(text)
        }
        # if choice was a dict turned to a string, restore it
        # also if the input and choice match if either the input or the choice is embedded in a language dict,
        # interpreting it as the english translation the corresponding choice is used
        for choice in schema.get('choices', []):
            if str(choice) == data['text'] or choice == {'en': data['text']} or data['text'] == {'en': choice}:
                data['text'] = choice
                break
    else:
        if not all(key.startswith(id_prefix + '__text_') for key in keys):
            raise ValueError('invalid text form data')
        if not keys:
            return None
        enabled_languages = form_data.get(id_prefix + '__text_languages', [])
        if 'en' not in enabled_languages and ('languages' not in schema or schema['languages'] == 'all' or 'en' in schema['languages']):
            enabled_languages.append('en')
        data = {
            '_type': 'text',
            'text': {}
        }
        for language in enabled_languages:
            text_list = form_data.get(id_prefix + '__text_' + language, [])
            text = text_list[0] if text_list else None
            if not text and not required:
                continue
            if text is None and schema.get('minLength', 0) > 0:
                return None
            if text is None:
                text = ""
            data['text'][language] = text
        if not data['text'] and not required:
            return None
    if schema.get('markdown'):
        data['is_markdown'] = True
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_hazards_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if id_prefix + '__hasnohazards' not in keys:
        raise ValueError(_('Please select at least one hazard or confirm that the object poses no hazards.'))
    hasnohazards_str = form_data.get(id_prefix + '__hasnohazards', [''])[0]
    if hasnohazards_str == 'true':
        hasnohazards = True
    elif hasnohazards_str == 'false':
        hasnohazards = False
    else:
        raise ValueError('invalid hazards form data')
    available_hazards = {
        'ghs01': 1, 'ghs02': 2, 'ghs03': 3, 'ghs04': 4, 'ghs05': 5, 'ghs06': 6, 'ghs07': 7, 'ghs08': 8, 'ghs09': 9
    }
    hazards = []
    for key in keys:
        if key == id_prefix + '__hasnohazards':
            continue
        if key == id_prefix + '__hidden':
            continue
        hazard_name = key.rsplit('__', 1)[1]
        if hazard_name in available_hazards:
            hazards.append(available_hazards[hazard_name])
        else:
            raise ValueError('invalid hazards form data')
    if hasnohazards != (len(hazards) == 0):
        raise ValueError(_('Please select at least one hazard or confirm that the object poses no hazards.'))
    data = {
        '_type': 'hazards',
        'hazards': hazards
    }
    return data


@form_data_parser
def parse_tags_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__tags']:
        raise ValueError('invalid tags form data')
    text = form_data.get(id_prefix + '__tags', [''])[0]
    tags = [tag.lower() for tag in text.split(',')]
    unique_tags = []
    for tag in tags:
        tag = ''.join(c for c in tag if not c.isspace())
        if not tag:
            continue
        if tag not in unique_tags:
            unique_tags.append(tag)
    tags = unique_tags
    data = {
        '_type': 'tags',
        'tags': tags
    }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_sample_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid sample form data')
    object_id_str = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please select a sample.'))
    try:
        object_id = int(object_id_str)
    except ValueError:
        raise ValueError('object_id must be int')
    if object_id == -1 and isinstance(previous_data, dict):
        data = previous_data
    else:
        data = {
            '_type': 'sample',
            'object_id': object_id
        }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_measurement_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid measurement form data')
    object_id_str = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please select a measurement.'))
    try:
        object_id = int(object_id_str)
    except ValueError:
        raise ValueError('object_id must be int')
    if object_id == -1 and isinstance(previous_data, dict):
        data = previous_data
    else:
        data = {
            '_type': 'measurement',
            'object_id': object_id
        }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_object_reference_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid object reference form data')
    object_id_str = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please select an object reference.'))
    try:
        object_id = int(object_id_str)
    except ValueError:
        raise ValueError('object_id must be int')
    if object_id == -1 and isinstance(previous_data, dict):
        data = previous_data
    else:
        data = {
            '_type': 'object_reference',
            'object_id': object_id
        }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_quantity_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if set(keys) != {id_prefix + '__magnitude', id_prefix + '__units'} and keys != [id_prefix + '__magnitude']:
        raise ValueError('invalid quantity form data')
    magnitude_str = form_data[id_prefix + '__magnitude'][0].strip()
    magnitude = None
    if not magnitude_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please enter a magnitude.'))

    if id_prefix + '__units' in form_data:
        units = form_data[id_prefix + '__units'][0]
        try:
            pint_units = ureg.Unit(units)
        except pint.UndefinedUnitError:
            raise ValueError('invalid units')
    else:
        units = '1'
        pint_units = ureg.Unit('1')
    dimensionality = str(int_ureg.Unit(pint_units).dimensionality)
    user_locale = languages.get_user_language(current_user).lang_code
    if units in ['min', 'h'] and ':' in magnitude_str:
        if units == 'min':
            match = re.match(re.compile(
                r'^(?P<minutes>[0-9]+):'
                r'(?P<seconds>[0-5][0-9]([' + babel.Locale(user_locale).number_symbols['latn']['decimal'] + '][0-9]+)?)$'
            ), magnitude_str)
        else:
            match = re.match(re.compile(
                r'^(?P<hours>([0-9]+)):'
                r'(?P<minutes>[0-5][0-9])'
                r'(:(?P<seconds>[0-5][0-9]([' + babel.Locale(user_locale).number_symbols['latn']['decimal'] + '][0-9]+)?))?$'
            ), magnitude_str)
        if match is None:
            raise ValueError(_('Unable to parse time.'))

        match_dict = match.groupdict()
        try:
            if 'hours' in match_dict.keys():
                hours = int(match_dict['hours'])
            else:
                hours = 0
            minutes = int(match_dict['minutes'])
            if match_dict['seconds']:
                seconds = parse_decimal(match_dict['seconds'], locale=user_locale, strict=True)
            else:
                seconds = decimal.Decimal(0)
        except ValueError:
            raise ValueError(_('Unable to parse time.'))
        magnitude_in_base_units = float(hours * 3600 + minutes * 60 + seconds)
        data = {
            '_type': 'quantity',
            'magnitude_in_base_units': magnitude_in_base_units,
            'dimensionality': dimensionality,
            'units': units
        }
    else:
        try:
            if magnitude is None and re.fullmatch('[-]?[0-9]+', magnitude_str):
                try:
                    magnitude = decimal.Decimal(int(magnitude_str))
                except ValueError:
                    pass
            if magnitude is None:
                try:
                    magnitude = parse_decimal(magnitude_str, locale=user_locale, strict=True)
                except Exception:
                    pass
            if magnitude is None:
                raise ValueError(_('Unable to parse magnitude.'))
        except ValueError:
            raise ValueError(_('The magnitude must be a number.'))
        magnitude_in_base_units = float(ureg.Quantity(magnitude, pint_units).to_base_units().magnitude)
        data = {
            '_type': 'quantity',
            'magnitude_in_base_units': magnitude_in_base_units,
            'magnitude': float(magnitude),
            'dimensionality': dimensionality,
            'units': units
        }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_datetime_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if keys != [id_prefix + '__datetime']:
        raise ValueError('invalid datetime form data')
    datetime_string = form_data.get(id_prefix + '__datetime', [''])[0]
    if not datetime_string:
        if not required:
            return None
        else:
            raise ValueError(_('Please enter a valid datetime.'))
    try:
        language = languages.get_user_language(current_user)
        parsed_datetime = datetime.datetime.strptime(datetime_string, language.datetime_format_datetime)
        # convert datetime to utc
        local_datetime = pytz.timezone(current_user.timezone or 'UTC').localize(parsed_datetime)
        utc_datetime = local_datetime.astimezone(pytz.utc)
        utc_datetime_str = utc_datetime.strftime('%Y-%m-%d %H:%M:%S')
    except Exception:
        raise ValueError(_('Please enter a valid datetime.'))
    data = {
        '_type': 'datetime',
        'utc_datetime': utc_datetime_str
    }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_boolean_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if set(keys) == {id_prefix + '__hidden', id_prefix + '__value'}:
        value = True
    elif keys == [id_prefix + '__hidden']:
        value = False
    else:
        raise ValueError('invalid boolean form data')
    return {
        '_type': 'bool',
        'value': value
    }


@form_data_parser
def parse_array_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str, errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.List[typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if not keys:
        return None
    item_schema = schema['items']
    item_indices = set()
    for key in keys:
        key_rest = key.replace(id_prefix, '', 1)
        if key_rest == '__hidden':
            continue
        if not key_rest.startswith('__'):
            raise ValueError('invalid array form data')
        key_parts = key_rest.split('__')
        if not key_parts[0] == '':
            raise ValueError('invalid array form data')
        try:
            item_index = int(key_parts[1])
        except ValueError:
            raise ValueError('invalid array form data')

        item_indices.add(item_index)
    items: typing.List[typing.Any] = []
    if item_indices:
        for i in item_indices:
            item_id_prefix = f'{id_prefix}__{i}'
            previous_item_data = previous_data[i] if isinstance(previous_data, list) and i < len(previous_data) else None
            items.append(parse_any_form_data(form_data, item_schema, item_id_prefix, errors, required=True, file_names_by_id=file_names_by_id, previous_data=previous_item_data))
    if not items and not required and schema.get('minItems', 0) > 0:
        return None
    schemas.validate(items, schema, strict=True, file_names_by_id=file_names_by_id)
    return items


@form_data_parser
def parse_object_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    assert schema['type'] == 'object'
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if not keys:
        return None
    data = {}
    for property_name, property_schema in schema['properties'].items():
        property_id_prefix = id_prefix + '__' + property_name
        property_required = property_name in schema.get('required', [])
        if isinstance(previous_data, dict) and property_name in previous_data:
            previous_property_data = previous_data[property_name]
        else:
            previous_property_data = None
        property = parse_any_form_data(form_data, property_schema, property_id_prefix, errors, required=property_required, file_names_by_id=file_names_by_id, previous_data=previous_property_data)
        if property is not None:
            data[property_name] = property
    schemas.validate(data, schema, strict=True, file_names_by_id=file_names_by_id)
    return data


@form_data_parser
def parse_user_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__uid']:
        raise ValueError('invalid user form data')
    user_id_str = form_data.get(id_prefix + '__uid', [''])[0]
    if not user_id_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please select a user.'))
    try:
        user_id = int(user_id_str)
    except ValueError:
        raise ValueError('user_id must be int')
    data = {
        '_type': 'user',
        'user_id': user_id
    }
    schemas.validate(data, schema, strict=True)
    return data


@form_data_parser
def parse_plotly_chart_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    plotly_chart_data_list = form_data.get(id_prefix + '__plotly', [])
    plotly_chart_data: typing.Optional[typing.Union[str, typing.Dict[str, typing.Any]]] = plotly_chart_data_list[0] if plotly_chart_data_list else None
    if plotly_chart_data is None and not required:
        plotly_chart_data = {}
    data = {
        '_type': 'plotly_chart',
        'plotly': plotly_chart_data
    }
    schemas.validate(data, schema, strict=True)

    return data


@form_data_parser
def parse_timeseries_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if set(keys) != {id_prefix + '__data', id_prefix + '__units'}:
        raise ValueError('invalid timeseries form data')
    units = form_data[id_prefix + '__units'][0]
    try:
        ureg.Unit(units)
    except pint.UndefinedUnitError:
        raise ValueError('invalid units')
    timeseries_data_str = form_data.get(id_prefix + '__data', [''])[0]
    if not timeseries_data_str and not required:
        return None
    timeseries_data: typing.List[typing.List[typing.Union[str, int, float]]]
    if not timeseries_data_str:
        timeseries_data = []
    else:
        csv_file = io.StringIO(timeseries_data_str)
        reader = csv.reader(csv_file, quoting=csv.QUOTE_NONNUMERIC)
        # csv.QUOTE_NONNUMERIC enables data conversion, however type annotations do not reflect this
        timeseries_data = list(reader)  # type: ignore
        if timeseries_data and all(isinstance(value, str) for value in timeseries_data[0]):
            # skip header rows
            timeseries_data = timeseries_data[1:]
        if not timeseries_data:
            raise ValueError(_('invalid timeseries CSV data, expected datetime string or relative time in seconds, magnitude and (optional) magnitude in base units'))
        is_relative_time = len(timeseries_data[0]) > 0 and type(timeseries_data[0][0]) in (int, float)
        user_timezone = pytz.timezone(current_user.timezone or 'UTC')
        for row in timeseries_data:
            if len(row) not in [2, 3] or not ((not is_relative_time and isinstance(row[0], str)) or (is_relative_time and type(row[0]) in (float, int))) or not all(type(entry) in (float, int) for entry in row[1:]):
                raise ValueError(_('invalid timeseries CSV data, expected datetime string or relative time in seconds, magnitude and (optional) magnitude in base units'))
            if not is_relative_time:
                try:
                    parsed_datetime = datetime.datetime.strptime(typing.cast(str, row[0]), '%Y-%m-%d %H:%M:%S.%f')
                except ValueError as exc:
                    raise ValueError(_('invalid datetime in timeseries, expected format: YYYY-MM-DD hh:mm:ss.ffffff')) from exc
                if user_timezone != pytz.utc:
                    # convert datetimes to UTC if necessary
                    try:
                        local_datetime = user_timezone.localize(parsed_datetime)
                        utc_datetime = local_datetime.astimezone(pytz.utc)
                    except Exception as exc:
                        raise ValueError(_('unable to convert datetime from your timezone to UTC')) from exc
                    row[0] = utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f')
    data = {
        '_type': 'timeseries',
        'data': timeseries_data,
        'units': units
    }
    schemas.validate(data, schema, strict=True)

    return data


@form_data_parser
def parse_file_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        id_prefix: str,
        errors: typing.Dict[str, str],
        *,
        file_names_by_id: typing.Dict[int, str],
        required: bool = False,
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__file_id']:
        raise ValueError('invalid file form data')
    file_id_str = form_data.get(id_prefix + '__file_id', [''])[0]
    if not file_id_str:
        if not required:
            return None
        else:
            raise ValueError(_('Please select a file.'))
    try:
        file_id = int(file_id_str)
    except ValueError:
        raise ValueError(_('Please select a file.'))
    if file_id not in file_names_by_id:
        raise ValueError(_('Please select a file.'))
    if 'extensions' in schema:
        file_name = file_names_by_id[file_id]
        if not any(file_name.lower().endswith(extension.lower()) for extension in schema['extensions']):
            raise ValueError(_('Please select a file with one of these extensions: %(extensions)s', extensions=', '.join(schema['extensions'])))
    data = {
        '_type': 'file',
        'file_id': file_id
    }
    schemas.validate(data, schema, strict=True, file_names_by_id=file_names_by_id)
    return data


def parse_form_data(
        form_data: typing.Dict[str, typing.List[str]],
        schema: typing.Dict[str, typing.Any],
        *,
        file_names_by_id: typing.Dict[int, str],
        previous_data: typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]] = None
) -> typing.Tuple[typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]], typing.Dict[str, str]]:
    id_prefix = 'object'
    errors: typing.Dict[str, str] = {}
    data = parse_object_form_data(form_data, schema, id_prefix, errors, file_names_by_id=file_names_by_id, previous_data=previous_data)
    return data, errors



# coding: utf-8
"""
Add missing NOT NULL constraint to column in markdown_images table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('markdown_images', 'utc_datetime'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE markdown_images
            ALTER COLUMN utc_datetime SET NOT NULL
    """))
    return True



# coding: utf-8
"""
Add enable_capacities to location_types table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_names = db.session.execute(db.text("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'location_types'
    """)).fetchall()
    if ('enable_capacities',) in column_names:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE location_types
            ADD enable_capacities BOOLEAN NOT NULL DEFAULT false
    """))
    db.session.execute(db.text("""
        ALTER TABLE location_types
            ALTER COLUMN enable_capacities DROP DEFAULT
    """))
    return True



# coding: utf-8
"""
Add component_id and id columns, fed key unique constraint and id primary key constraint to markdown_images table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('markdown_images', 'component_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE markdown_images
            ADD component_id INTEGER,
            ADD FOREIGN KEY(component_id) REFERENCES components(id),
            DROP CONSTRAINT markdown_images_pkey,
            ADD COLUMN id SERIAL PRIMARY KEY,
            ADD CONSTRAINT markdown_images_file_name_component_id_key UNIQUE(file_name, component_id)
    """))
    return True



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .languages import Language

__author__ = 'Du Kim Nguyen <k.nguyen@fz-juelich.de>'


class InstrumentTranslation(Model):
    __tablename__ = 'instrument_translations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)

    instrument_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('instruments.id'), nullable=False)

    language_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('languages.id'), nullable=False)
    language: Mapped['Language'] = relationship('Language')

    name: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')
    description: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')
    notes: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')
    short_description: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentTranslation"]]

    __table_args__ = (
        db.UniqueConstraint('language_id', 'instrument_id', name='_language_id_instrument_id_uc'),
        db.CheckConstraint(
            'NOT (name IS NULL AND description IS NULL AND notes IS NULL AND short_description IS NULL)',
            name='instrument_translations_not_empty_check'
        )
    )

    def __init__(
            self,
            instrument_id: int,
            language_id: int,
            name: str,
            description: str = '',
            notes: str = '',
            short_description: str = ''
    ) -> None:
        super().__init__(
            instrument_id=instrument_id,
            language_id=language_id,
            name=name,
            description=description,
            notes=notes,
            short_description=short_description
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, InstrumentTranslation):
            return bool(
                self.id == other.id and
                self.language_id == other.language_id and
                self.instrument_id == other.instrument_id and
                self.name == other.name and
                self.description == other.description and
                self.notes == other.notes and
                self.short_description == other.short_description
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, name={self.language_id})>'



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .groups import Group
from .users import User
from .actions import Action
from .permissions import Permissions
from .projects import Project
from .utils import Model

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserActionPermissions(Model):
    __tablename__ = 'user_action_permissions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserActionPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, user_id),
    )


class GroupActionPermissions(Model):
    __tablename__ = 'group_action_permissions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupActionPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, group_id),
    )


class ProjectActionPermissions(Model):
    __tablename__ = 'project_action_permissions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ProjectActionPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, project_id),
    )


class AllUserActionPermissions(Model):
    __tablename__ = 'all_user_action_permissions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Action.id), primary_key=True)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["AllUserActionPermissions"]]



# coding: utf-8
"""
This file should only contain the definition of the __version__ variable.

It may be read by multiple other files, e.g. setup.py and the status page.

The SAMPLEDB_VERSION environment variable can be used to override the static
value. This is done when building the Docker image as part of the GitLab CI
pipeline to provide the current version as returned by git describe.
"""

import os

__version__ = os.environ.get('SAMPLEDB_VERSION') or '0.32.0'



# coding: utf-8
"""

"""

import flask
import flask_login

from . import frontend
from ..logic.publications import get_publications_for_user
from ..utils import FlaskResponseT


@frontend.route('/publications/')
@flask_login.login_required
def publications() -> FlaskResponseT:
    return flask.render_template(
        'publications.html',
        publications=get_publications_for_user(flask_login.current_user.id)
    )



# coding: utf-8
"""

"""
import flask_login
import flask_wtf
import wtforms.fields
import wtforms.validators
import pytz

from . import frontend
from ..logic import settings
from ..utils import FlaskResponseT


class TimezoneForm(flask_wtf.FlaskForm):
    timezone = wtforms.fields.StringField(
        validators=[
            wtforms.validators.InputRequired()
        ]
    )

    def validate_timezone(self, field: wtforms.fields.StringField) -> None:
        try:
            pytz.timezone(field.data)
        except Exception:
            raise wtforms.validators.ValidationError("unknown time zone")


@frontend.route('/set-timezone', methods=["POST"])
@flask_login.login_required
def set_timezone() -> FlaskResponseT:
    if not settings.get_user_setting(flask_login.current_user.id, 'AUTO_TZ'):
        return '', 200
    form = TimezoneForm()
    if form.validate_on_submit():
        settings.set_user_settings(flask_login.current_user.id, {
            'TIMEZONE': form.timezone.data
        })
        return '', 200
    return '', 400



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


class Tag(Model):
    __tablename__ = 'tags'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[str] = db.Column(db.String, unique=True, nullable=False)
    uses: Mapped[int] = db.Column(db.Integer, nullable=False, default=0)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Tag"]]



from datetime import datetime
import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_bool, _get_utc_datetime, _get_translation, _get_dict
from .users import _parse_user_ref, _get_or_create_user_id, UserRef
from .locations import _get_or_create_location_id, _parse_location_ref, LocationRef
from ..locations import create_fed_assignment, get_fed_object_location_assignment, ObjectLocationAssignment, _get_mutable_object_location_assignment
from ..components import Component
from .. import errors, fed_logs, object_log
from ...models import Object
from ... import db


class ObjectLocationAssignmentData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    location: typing.Optional[LocationRef]
    responsible_user: typing.Optional[UserRef]
    user: typing.Optional[UserRef]
    description: typing.Optional[typing.Dict[str, str]]
    utc_datetime: datetime
    confirmed: bool
    declined: bool


def import_object_location_assignment(
        assignment_data: ObjectLocationAssignmentData,
        object: Object,
        component: Component
) -> ObjectLocationAssignment:
    component_id = _get_or_create_component_id(assignment_data['component_uuid'])
    if component_id is not None:
        assignment = get_fed_object_location_assignment(assignment_data['fed_id'], component_id)
    else:
        assignment = _get_mutable_object_location_assignment(assignment_data['fed_id'])

    user_id = _get_or_create_user_id(assignment_data['user'])
    responsible_user_id = _get_or_create_user_id(assignment_data['responsible_user'])
    location_id = _get_or_create_location_id(assignment_data['location'])

    if assignment is None:
        assert component_id is not None
        assignment = create_fed_assignment(assignment_data['fed_id'], component_id, object.object_id, location_id, responsible_user_id, user_id, assignment_data['description'], assignment_data['utc_datetime'], assignment_data['confirmed'], assignment_data.get('declined', False))
        fed_logs.import_object_location_assignment(assignment.id, component.id)
        if user_id:
            object_log.assign_location(user_id, object_id=object.object_id, object_location_assignment_id=assignment.id, utc_datetime=assignment.utc_datetime, is_imported=True)
    elif assignment.location_id != location_id or assignment.user_id != user_id or assignment.responsible_user_id != responsible_user_id or assignment.description != assignment_data['description'] or assignment.object_id != object.object_id or assignment.confirmed != assignment_data['confirmed'] or assignment.declined != assignment_data.get('declined', False) or assignment.utc_datetime != assignment_data['utc_datetime']:
        assignment.location_id = location_id
        assignment.responsible_user_id = responsible_user_id
        assignment.user_id = user_id
        assignment.description = assignment_data['description']
        assignment.object_id = object.object_id
        assignment.confirmed = assignment_data['confirmed']
        assignment.utc_datetime = assignment_data['utc_datetime']
        assignment.declined = assignment_data.get('declined', False)
        db.session.commit()
        fed_logs.update_object_location_assignment(assignment.id, component.id)
    return ObjectLocationAssignment.from_database(assignment)


def parse_object_location_assignment(
        assignment_data: typing.Dict[str, typing.Any]
) -> ObjectLocationAssignmentData:
    uuid = _get_uuid(assignment_data.get('component_uuid'))
    fed_id = _get_id(assignment_data.get('id'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local object location assignment {fed_id} @ {uuid}')

    responsible_user_data = _get_dict(assignment_data.get('responsible_user'))
    location_data = _get_dict(assignment_data.get('location'))
    description = _get_translation(assignment_data.get('description'))
    if responsible_user_data is None and location_data is None and description is None:
        raise errors.InvalidDataExportError(f'Empty object location assignment {fed_id} @ {uuid}')

    return ObjectLocationAssignmentData(
        fed_id=fed_id,
        component_uuid=uuid,
        location=_parse_location_ref(location_data),
        responsible_user=_parse_user_ref(responsible_user_data),
        user=_parse_user_ref(_get_dict(assignment_data.get('user'))),
        description=description,
        utc_datetime=_get_utc_datetime(assignment_data.get('utc_datetime'), mandatory=True),
        confirmed=_get_bool(assignment_data.get('confirmed'), default=False),
        declined=_get_bool(assignment_data.get('declined'), default=False)
    )


def parse_import_object_location_assignment(
        assignment_data: typing.Dict[str, typing.Any],
        object: Object,
        component: Component
) -> ObjectLocationAssignment:
    return import_object_location_assignment(parse_object_location_assignment(assignment_data), object, component)



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .languages import Language


class ActionTypeTranslation(Model):
    __tablename__ = 'action_type_translations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)

    language_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('languages.id'), nullable=False)
    language: Mapped['Language'] = relationship('Language')

    action_type_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('action_types.id'), nullable=False)

    name: Mapped[str] = db.Column(db.String, nullable=False)
    description: Mapped[str] = db.Column(db.String, nullable=False)
    object_name: Mapped[str] = db.Column(db.String, nullable=False)
    object_name_plural: Mapped[str] = db.Column(db.String, nullable=False)
    view_text: Mapped[str] = db.Column(db.String, nullable=False)
    perform_text: Mapped[str] = db.Column(db.String, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ActionTypeTranslation"]]

    __table_args__ = (
        db.UniqueConstraint('language_id', 'action_type_id', name='_language_id_action_type_id_uc'),
        # TODO: action translation and instrument translation have not-null contraints, but this doesn't
    )

    def __init__(
            self,
            language_id: int,
            action_type_id: int,
            name: str,
            description: str,
            object_name: str,
            object_name_plural: str,
            view_text: str,
            perform_text: str
    ) -> None:
        super().__init__(
            language_id=language_id,
            action_type_id=action_type_id,
            name=name,
            description=description,
            object_name=object_name,
            object_name_plural=object_name_plural,
            view_text=view_text,
            perform_text=perform_text
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, ActionTypeTranslation):
            return bool(
                self.id == other.id and
                self.language_id == other.language_id and
                self.action_type_id == other.action_type_id and
                self.name == other.name and
                self.description == other.description and
                self.object_name == other.object_name and
                self.object_name_plural == other.object_name_plural and
                self.view_text == other.view_text and
                self.perform_text == other.perform_text
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id!r}, name={self.name!r})>'


class ActionTranslation(Model):
    __tablename__ = 'action_translations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)

    language_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('languages.id'), nullable=False)
    language: Mapped['Language'] = relationship('Language')

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('actions.id'), nullable=False)

    name: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    description: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')
    short_description: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True, default='')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ActionTranslation"]]

    __table_args__ = (
        db.UniqueConstraint('language_id', 'action_id', name='_language_id_action_id_uc'),
        db.CheckConstraint(
            'NOT (name IS NULL AND description IS NULL AND short_description IS NULL)',
            name='action_translations_not_empty_check'
        )
    )

    def __init__(
            self,
            action_id: int,
            language_id: int,
            name: str,
            description: str = '',
            short_description: str = '',
    ) -> None:
        super().__init__(
            action_id=action_id,
            language_id=language_id,
            name=name,
            description=description,
            short_description=short_description
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, ActionTranslation):
            return bool(
                self.id == other.id and
                self.action_id == other.action_id and
                self.language_id == other.language_id and
                self.name == other.name and
                self.description == other.description and
                self.short_description == other.short_description
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id!r})>'




import datetime
import string
import typing
import uuid
import urllib.parse

import flask
import requests

from .objects import get_object
from .object_permissions import get_user_object_permissions
from .datatypes import DateTime, Quantity, Timeseries
from . import settings, users, errors, utils, files
from ..models import SciCatExport, SciCatExportType, Permissions
from .. import db


SCICAT_TIMEOUT = 30


def get_scicat_url(
        object_id: int
) -> typing.Optional[str]:
    """
    Return the URL for an object in a SciCat instance.

    :param object_id: the ID of an existing object
    :return: the URL or None
    """
    scicat_export: typing.Optional[SciCatExport] = SciCatExport.query.filter_by(object_id=object_id).first()
    if scicat_export is None:
        return None
    return scicat_export.scicat_url


def _convert_metadata(
        data: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        schema: typing.Dict[str, typing.Any],
        property_whitelist: typing.List[typing.List[typing.Union[str, int]]],
        user_id: int,
        object_id: int
) -> typing.Optional[typing.Union[str, float, int, typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
    if property_whitelist is None:
        property_whitelist = [['name']]

    def _convert_object(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        property_whitelist: typing.List[typing.List[typing.Union[str, int]]],
        object_id: int
    ) -> typing.Dict[str, typing.Any]:
        metadata = {}
        for property_name in schema['properties']:
            if property_name not in data:
                continue
            if not all(c in string.ascii_letters + string.digits + '_' for c in property_name):
                # skip invalid property names
                continue
            sub_property_whitelist = []
            for path in property_whitelist:
                if path and path[0] == property_name:
                    sub_property_whitelist.append(path[1:])
            if not sub_property_whitelist:
                continue
            property_data = data[property_name]
            property_schema = schema['properties'][property_name]
            converted_data = _convert_metadata(property_data, property_schema, sub_property_whitelist, user_id, object_id)
            if converted_data is not None:
                metadata[property_name] = converted_data
        return metadata

    def _convert_array(
        data: typing.List[typing.Any],
        schema: typing.Dict[str, typing.Any],
        property_whitelist: typing.List[typing.List[typing.Union[str, int]]],
        object_id: int
    ) -> typing.List[typing.Any]:
        item_schema = schema['items']
        metadata = []
        for index, item_data in enumerate(data):
            sub_property_whitelist = []
            for path in property_whitelist:
                if path and path[0] == index:
                    sub_property_whitelist.append(path[1:])
            if not sub_property_whitelist:
                continue
            converted_data = _convert_metadata(item_data, item_schema, sub_property_whitelist, user_id, object_id)
            if converted_data is not None:
                metadata.append(converted_data)
        return metadata

    def _convert_timeseries(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> typing.List[typing.Any]:
        units = data['units']
        rows = data['data']
        metadata = []
        for row in rows:
            utc_datetime_str, magnitude = row[:2]
            utc_datetime_str = datetime.datetime.strptime(utc_datetime_str, Timeseries.DATETIME_FORMAT_STRING).isoformat(timespec='microseconds')
            if units in {'1', ''}:
                value = magnitude
            else:
                value = {
                    'value': str(float(round(magnitude, 13))),
                    'unit': units
                }
            metadata.append({
                'utc_datetime': utc_datetime_str,
                'value': value
            })
        return metadata

    def _convert_datetime(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> str:
        return datetime.datetime.strptime(data['utc_datetime'], DateTime.FORMAT_STRING).isoformat(timespec='microseconds')

    def _convert_bool(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> bool:
        return bool(data['value'])

    def _convert_text(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> str:
        if isinstance(data['text'], dict):
            return str(data['text'].get('en', data['text']))
        if isinstance(data['text'], str):
            return data['text']
        return ''

    def _convert_quantity(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> typing.Union[float, int, typing.Dict[str, typing.Any]]:
        if data['units'] in {'1', ''}:
            return typing.cast(typing.Union[float, int], data['magnitude_in_base_units'])
        quantity = Quantity.from_json(data)
        return {
            'value': str(float(round(quantity.magnitude, 13))),
            'unit': quantity.units
        }

    def _convert_object_reference(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> str:
        object_id = data['object_id']
        if Permissions.READ in get_user_object_permissions(object_id=object_id, user_id=user_id):
            raw_object_name = get_object(object_id).name
        else:
            raw_object_name = None
        object_name = utils.get_translated_text(raw_object_name, 'en', f"Object #{object_id}")
        sampledb_object_url = flask.url_for(
            'frontend.object',
            object_id=object_id,
            _external=True
        )
        scicat_export = get_scicat_export_for_object(object_id)
        if scicat_export is None:
            return f"{object_name} ({sampledb_object_url})"
        else:
            return f"{object_name} ({sampledb_object_url} / {scicat_export.scicat_url})"

    def _convert_user(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> str:
        try:
            user = users.get_user(data['user_id'])
            return user.get_name(include_ref=True)
        except errors.UserDoesNotExistError:
            return f"User #{data['user_id']}"

    def _convert_file(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        object_id: int
    ) -> str:
        try:
            file = files.get_file(data['file_id'], object_id)
            if file.is_hidden:
                return f'Hidden (#{data["file_id"]})'
            else:
                return file.original_file_name
        except errors.FileDoesNotExistError:
            return f'Unknown (#{data["file_id"]})'
        except errors.InvalidFileStorageError:
            return f'Unnamed (#{data["file_id"]})'

    def _convert_hazards(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
    ) -> str:
        hazard_names = {
            1: 'Explosive',
            2: 'Flammable',
            3: 'Oxidizing',
            4: 'Compressed Gas',
            5: 'Corrosive',
            6: 'Toxic',
            7: 'Harmful',
            8: 'Health Hazard',
            9: 'Environmental Hazard'
        }
        return ', '.join(hazard_names[hazard_id] for hazard_id in sorted(data['hazards']))

    if schema['type'] == 'object' and isinstance(data, dict):
        return _convert_object(data, schema, property_whitelist, object_id)
    if schema['type'] == 'array' and isinstance(data, list):
        return _convert_array(data, schema, property_whitelist, object_id)
    if schema['type'] == 'datetime' and isinstance(data, dict):
        return _convert_datetime(data, schema)
    if schema['type'] == 'bool' and isinstance(data, dict):
        return _convert_bool(data, schema)
    if schema['type'] == 'text' and isinstance(data, dict):
        return _convert_text(data, schema)
    if schema['type'] == 'quantity' and isinstance(data, dict):
        return _convert_quantity(data, schema)
    if schema['type'] in {'sample', 'measurement', 'object_reference'} and isinstance(data, dict):
        return _convert_object_reference(data, schema)
    if schema['type'] == 'user' and isinstance(data, dict):
        return _convert_user(data, schema)
    if schema['type'] == 'hazards' and isinstance(data, dict):
        return _convert_hazards(data, schema)
    if schema['type'] == 'timeseries' and isinstance(data, dict):
        return _convert_timeseries(data, schema)
    if schema['type'] == 'file' and isinstance(data, dict):
        return _convert_file(data, schema, object_id)
    if schema['type'] == 'tags':
        # tags are exported as keywords
        return None
    return None


def upload_object(
        object_id: int,
        user_id: int,
        api_url: str,
        frontend_url: str,
        api_token: str,
        property_whitelist: typing.List[typing.List[typing.Union[str, int]]],
        tag_whitelist: typing.List[str],
        object_export_type: SciCatExportType,
        owner_group: str,
        access_groups: typing.Optional[typing.List[str]] = None,
        instrument_pid: typing.Optional[str] = None,
        sample_pid: typing.Optional[str] = None,
        input_dataset_pids: typing.Optional[typing.List[str]] = None
) -> typing.Optional[str]:
    user = users.get_user(user_id)
    object = get_object(object_id)
    original_object_version = get_object(object_id, version_id=0)

    object_url = flask.url_for('frontend.object', object_id=object_id, _external=True)
    pid_prefix = flask.current_app.config["SCICAT_EXTRA_PID_PREFIX"]
    pid_uuid = uuid.uuid4()
    pid = f"{pid_prefix}{pid_uuid}"

    object_name = utils.get_translated_text(object.name, 'en', default=f"Object #{object_id}")

    if object_export_type in {SciCatExportType.RAW_DATASET, SciCatExportType.DERIVED_DATASET}:
        if object_export_type == SciCatExportType.RAW_DATASET:
            api_endpoint = f'{api_url}/api/v3/RawDatasets'
        else:
            api_endpoint = f'{api_url}/api/v3/DerivedDatasets'
        scicat_url = f'{frontend_url}/datasets/{urllib.parse.quote_plus("PID/" + pid)}'
        data: typing.Dict[str, typing.Any] = {
            "pid": pid,
            "datasetName": object_name,
            "owner": user.name,
            "ownerEmail": user.email,
            "ownerGroup": owner_group,
            "contactEmail": user.email,
            "creationTime": original_object_version.utc_datetime.strftime('%Y-%m-%d %H:%M:%S') if original_object_version.utc_datetime else None,
            "description": f"Metadata export from {flask.current_app.config['SERVICE_NAME']} object #{object_id}",
            "sourceFolder": "-",
            "version": str(object.version_id),
            "keywords": [],
            "scientificMetadata": _convert_metadata(object.data, object.schema, property_whitelist, user_id, object.object_id) if object.data is not None and object.schema is not None else {}
        }
        if object_export_type == SciCatExportType.RAW_DATASET:
            data.update({
                "principalInvestigator": "-",
                "creationLocation": object_url,
            })
            if instrument_pid is not None:
                data["instrumentId"] = instrument_pid
            if sample_pid is not None:
                data["sampleId"] = sample_pid
        else:
            if not input_dataset_pids:
                input_dataset_pids = ["-"]
            data.update({
                "investigator": "-",
                "usedSoftware": ["-"],
                "inputDatasets": input_dataset_pids
            })

        if object.schema is not None and object.data is not None and 'tags' in object.schema['properties'] and object.schema['properties']['tags']['type'] == 'tags' and 'tags' in object.data:
            for tag in object.data['tags']['tags']:
                if tag in tag_whitelist:
                    data['keywords'].append(tag)

        if user.orcid:
            data["orcidOfOwner"] = user.orcid

        if access_groups:
            data["accessGroups"] = access_groups
    elif object_export_type == SciCatExportType.SAMPLE:
        api_endpoint = f'{api_url}/api/v3/Samples'
        scicat_url = f'{frontend_url}/samples/{urllib.parse.quote(object_name + " / " + pid, safe="")}'
        data = {
            "sampleId": object_name + " / " + pid,
            "owner": user.name,
            "description": f"Metadata export from {flask.current_app.config['SERVICE_NAME']} object #{object_id}",
            "sampleCharacteristics": _convert_metadata(object.data, object.schema, property_whitelist, user_id, object.object_id) if object.data is not None and object.schema is not None else {},
            "ownerGroup": owner_group
        }

        if access_groups:
            data["accessGroups"] = access_groups
    else:
        return None

    try:
        r = requests.post(
            url=api_endpoint,
            json=[data],
            headers={
                'Authorization': api_token
            },
            timeout=SCICAT_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.SciCatNotReachableError()
    if r.status_code != 200:
        return None

    db.session.add(SciCatExport(object_id, scicat_url, pid, user_id, object_export_type))
    db.session.commit()
    return scicat_url


def get_user_valid_api_token(api_url: str, user_id: int) -> typing.Optional[str]:
    """
    Return whether a user has a valid API token for a given SciCat instance.

    :param api_url: the base URL of the SciCat API server
    :param user_id: the ID of an existing user
    :return: a valid API token stored for the user, or None
    :raise errors.UserDoesNotExistError: if not user with the given user ID
        exists
    :raise errors.SciCatNotReachableError: if there was an error during
        communication with the SciCat API
    """
    api_token: typing.Optional[str] = settings.get_user_setting(user_id, 'SCICAT_API_TOKEN')
    if not api_token:
        # make sure the user even exists
        users.check_user_exists(user_id)
        return None
    if not is_api_token_valid(api_url, api_token):
        return None
    return api_token


def is_api_token_valid(api_url: str, api_token: str) -> bool:
    """
    Return whether an API token is valid for a given SciCat instance.

    :param api_url: the base URL of the SciCat API server
    :param api_token: the API token to validate
    :return: whether the API token is valid
    :raise errors.SciCatNotReachableError: if there was an error during
        communication with the SciCat API
    """
    # check basic API token structure
    if len(api_token) != 64:
        return False
    if not all(c in (string.ascii_letters + string.digits) for c in api_token):
        return False
    # check whether the API token can actually be used
    try:
        r = requests.get(
            f'{api_url}/api/v3/Datasets/count',
            headers={
                'Authorization': api_token
            },
            allow_redirects=False,
            timeout=SCICAT_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.SciCatNotReachableError()
    if r.status_code != 200:
        return False
    return True


def get_property_export_default(
        path: typing.List[str],
        schema: typing.Dict[str, typing.Any]
) -> bool:
    """
    Return whether scicat_export is set to True in the schema of a property.

    :param path: the path to a property
    :param schema: the object schema
    :return: the value of scicat_export, or False if it is not set
    """
    if path == ['name']:
        return True
    try:
        subschema = schema
        for key in path:
            if isinstance(key, str):
                subschema = subschema['properties'][key]
            else:
                subschema = subschema['items']
        return bool(subschema.get('scicat_export', False))
        # TODO: allow scicat_export in schemas
    except Exception:
        return False


def get_user_groups(
        api_url: str,
        api_token: str
) -> typing.List[str]:
    """
    Return the list of SciCat groups the current user is in.

    :param api_url: the base URL of the SciCat API server
    :param api_token: the API token to validate
    :return: the list of group names
    :raise errors.SciCatNotReachableError: if there was an error during
        communication with the SciCat API
    """
    try:
        r = requests.get(
            f'{api_url}/api/v3/Users/userInfos',
            headers={
                'Authorization': api_token
            },
            allow_redirects=False,
            timeout=SCICAT_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.SciCatNotReachableError()
    if r.status_code != 200:
        return []
    try:
        unfiltered_groups = r.json().get('currentGroups')
    except Exception:
        return []
    # remove duplicates, but keep order
    user_groups = []
    for user_group in unfiltered_groups:
        if user_group not in user_groups:
            user_groups.append(user_group)
    return user_groups


def get_instruments(
        api_url: str,
        api_token: str
) -> typing.List[typing.Tuple[str, str]]:
    """
    Return the list of SciCat instruments.

    :param api_url: the base URL of the SciCat API server
    :param api_token: the API token to validate
    :return: the list of instruments as tuples containing pid and name
    :raise errors.SciCatNotReachableError: if there was an error during
        communication with the SciCat API
    """
    try:
        r = requests.get(
            f'{api_url}/api/v3/Instruments',
            headers={
                'Authorization': api_token
            },
            allow_redirects=False,
            timeout=SCICAT_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.SciCatNotReachableError()
    if r.status_code != 200:
        return []
    return [
        (instrument['pid'], instrument['name'])
        for instrument in r.json()
    ]


def get_scicat_export_for_object(
        object_id: int
) -> typing.Optional[SciCatExport]:
    return SciCatExport.query.filter_by(object_id=object_id).first()


def get_exported_referenced_objects(
        data: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        schema: typing.Dict[str, typing.Any],
        user_id: int
) -> typing.Set[SciCatExport]:
    exports: typing.Set[SciCatExport] = set()
    if schema['type'] == 'object' and isinstance(data, dict):
        for property_name in schema['properties']:
            if property_name not in data:
                continue
            property_data = data[property_name]
            property_schema = schema['properties'][property_name]
            exports.update(get_exported_referenced_objects(property_data, property_schema, user_id))
    if schema['type'] == 'array' and isinstance(data, list):
        item_schema = schema['items']
        for item_data in data:
            exports.update(get_exported_referenced_objects(item_data, item_schema, user_id))
    if schema['type'] in {'sample', 'measurement', 'object_reference'} and isinstance(data, dict):
        export = get_scicat_export_for_object(data['object_id'])
        if export is not None:
            exports.add(export)
    return exports



# coding: utf-8
"""

"""

import datetime
import typing

import flask

from . import object_permissions, errors
from . import users
from .background_tasks.trigger_webhooks import post_trigger_object_log_webhooks
from .users import get_user
from ..models import ObjectLogEntry, ObjectLogEntryType, Permissions, Objects
from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def object_log_entry_to_json(log_entry: ObjectLogEntry) -> typing.Dict[str, typing.Any]:
    return {
        'log_entry_id': log_entry.id,
        'type': log_entry.type.name,
        'object_id': log_entry.object_id,
        'user_id': log_entry.user_id,
        'data': {key: value for key, value in log_entry.data.items() if key not in ['object', 'sample', 'measurement']},
        'utc_datetime': log_entry.utc_datetime.strftime('%Y-%m-%d %H:%M:%S'),
        'is_imported': log_entry.is_imported
    }


def get_object_log_entries_by_user(user_id: int, after_id: int = 0) -> typing.List[ObjectLogEntry]:
    user = get_user(user_id)
    if user.has_admin_permissions:
        object_log_entries = db.session.execute(db.select(ObjectLogEntry).where(ObjectLogEntry.id > after_id).order_by(db.desc(ObjectLogEntry.utc_datetime))).all()
    else:
        stmt = db.text("""
            SELECT DISTINCT object_id
            FROM user_object_permissions_by_all
            WHERE (user_id = :user_id OR user_id IS NULL) AND (requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (requires_instruments IS FALSE OR :enable_instruments IS TRUE)
        """).columns(
            Objects._current_table.c.object_id,
        ).subquery('readable_objects')
        stmt = db.select(ObjectLogEntry).join(stmt).where(ObjectLogEntry.id > after_id).order_by(db.desc(ObjectLogEntry.utc_datetime))

        object_log_entries = db.session.execute(stmt, {
            'user_id': user_id,
            'enable_anonymous_users': flask.current_app.config['ENABLE_ANONYMOUS_USERS'],
            'enable_instruments': not flask.current_app.config['DISABLE_INSTRUMENTS'],
            'after_id': after_id
        }).all()
    return process_object_log_entries([row[0] for row in object_log_entries], user_id)


def process_object_log_entries(object_log_entries: typing.List[ObjectLogEntry], user_id: typing.Optional[int] = None) -> typing.List[ObjectLogEntry]:
    processed_object_log_entries = []
    users_by_id: typing.Dict[typing.Optional[int], typing.Optional[users.User]] = {None: None}
    referenced_object_ids = {}
    referenced_object_type_by_log_entry_type = {
        ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT: 'measurement',
        ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION: 'sample',
        ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA: 'object'
    }
    for object_log_entry in object_log_entries:
        using_object_type = referenced_object_type_by_log_entry_type.get(object_log_entry.type)
        if using_object_type is not None:
            using_object_id = using_object_type + '_id'
            object_id_value = object_log_entry.data.get(using_object_id)
            if type(object_id_value) is int:
                referenced_object_ids[object_log_entry.id] = object_id_value
        if object_log_entry.user_id not in users_by_id:
            users_by_id[object_log_entry.user_id] = users.get_user(object_log_entry.user_id)
        object_log_entry.user = users_by_id[object_log_entry.user_id]
        # remove the modified log entry from the session to avoid an
        # accidental commit of these changes
        db.session.expunge(object_log_entry)
        processed_object_log_entries.append(object_log_entry)
    if referenced_object_ids:
        referenced_objects = object_permissions.get_objects_with_permissions(
            user_id=user_id,
            permissions=Permissions.READ,
            object_ids=list(set(referenced_object_ids.values()))
        )
        referenced_objects_by_id = {
            object.id: object
            for object in referenced_objects
        }
        for object_log_entry in processed_object_log_entries:
            if object_log_entry.id in referenced_object_ids:
                using_object_type = referenced_object_type_by_log_entry_type[object_log_entry.type]
                using_object_id = using_object_type + '_id'
                referenced_object_id = referenced_object_ids[object_log_entry.id]
                if referenced_object_id in referenced_objects_by_id:
                    referenced_object = referenced_objects_by_id[referenced_object_id]
                    object_log_entry.data[using_object_type] = referenced_object
                else:
                    # Clear the using object ID, the user may only know that the
                    # object was used for some other object, but not for which
                    object_log_entry.data[using_object_id] = None
    return processed_object_log_entries


def get_object_log_entry(object_log_entry_id: int, user_id: typing.Optional[int] = None) -> ObjectLogEntry:
    object_log_entry = ObjectLogEntry.query.filter_by(id=object_log_entry_id).order_by(db.desc(ObjectLogEntry.utc_datetime)).first()
    if object_log_entry is None:
        raise errors.ObjectLogEntryDoesNotExistError()
    return process_object_log_entries([object_log_entry], user_id)[0]


def get_object_log_entries(object_id: int, user_id: typing.Optional[int] = None) -> typing.List[ObjectLogEntry]:
    object_log_entries = ObjectLogEntry.query.filter_by(object_id=object_id).order_by(db.desc(ObjectLogEntry.utc_datetime)).all()
    return process_object_log_entries(object_log_entries, user_id)


def _store_new_log_entry(
        type: ObjectLogEntryType,
        object_id: int,
        user_id: int,
        data: typing.Dict[str, typing.Any],
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    object_log_entry = ObjectLogEntry(
        type=type,
        object_id=object_id,
        user_id=user_id,
        data=data,
        utc_datetime=datetime.datetime.now(datetime.timezone.utc) if utc_datetime is None else utc_datetime,
        is_imported=is_imported
    )
    db.session.add(object_log_entry)
    db.session.commit()
    post_trigger_object_log_webhooks(object_log_entry)


def create_object(
        user_id: int,
        object_id: int,
        previous_object_id: typing.Optional[int] = None,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    data = {}
    if previous_object_id:
        data['previous_object_id'] = previous_object_id
    _store_new_log_entry(
        type=ObjectLogEntryType.CREATE_OBJECT,
        object_id=object_id,
        user_id=user_id,
        data=data,
        utc_datetime=utc_datetime,
        is_imported=is_imported
    )


def edit_object(
        user_id: int,
        object_id: int,
        version_id: int,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.EDIT_OBJECT,
        object_id=object_id,
        user_id=user_id,
        data={
            'version_id': version_id
        },
        utc_datetime=utc_datetime,
        is_imported=is_imported
    )


def restore_object_version(user_id: int, object_id: int, version_id: int, restored_version_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.RESTORE_OBJECT_VERSION,
        object_id=object_id,
        user_id=user_id,
        data={
            'version_id': version_id,
            'restored_version_id': restored_version_id
        }
    )


def use_object_in_measurement(user_id: int, object_id: int, measurement_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT,
        object_id=object_id,
        user_id=user_id,
        data={
            'measurement_id': measurement_id
        }
    )


def use_object_in_sample(user_id: int, object_id: int, sample_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'sample_id': sample_id
        }
    )


def post_comment(
        user_id: int,
        object_id: int,
        comment_id: int,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.POST_COMMENT,
        object_id=object_id,
        user_id=user_id,
        data={
            'comment_id': comment_id
        },
        utc_datetime=utc_datetime,
        is_imported=is_imported
    )


def upload_file(
        user_id: int,
        object_id: int,
        file_id: int,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.UPLOAD_FILE,
        object_id=object_id,
        user_id=user_id,
        data={
            'file_id': file_id
        },
        utc_datetime=utc_datetime,
        is_imported=is_imported
    )


def create_batch(user_id: int, object_id: int, batch_object_ids: typing.List[int]) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.CREATE_BATCH,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_ids': batch_object_ids
        }
    )


def assign_location(
        user_id: int,
        object_id: int,
        object_location_assignment_id: int,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        is_imported: bool = False
) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.ASSIGN_LOCATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        },
        utc_datetime=utc_datetime,
        is_imported=is_imported
    )


def link_publication(
        user_id: int,
        object_id: int,
        doi: str,
        title: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None
) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.LINK_PUBLICATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'doi': doi,
            'title': title,
            'object_name': object_name
        }
    )


def reference_object_in_metadata(user_id: int, object_id: int, referencing_object_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_id': referencing_object_id
        }
    )


def export_to_dataverse(user_id: int, object_id: int, dataverse_url: str) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.EXPORT_TO_DATAVERSE,
        object_id=object_id,
        user_id=user_id,
        data={
            'dataverse_url': dataverse_url
        }
    )


def link_project(user_id: int, object_id: int, project_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.LINK_PROJECT,
        object_id=object_id,
        user_id=user_id,
        data={
            'project_id': project_id
        }
    )


def unlink_project(user_id: int, object_id: int, project_id: int, project_deleted: bool = False) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.UNLINK_PROJECT,
        object_id=object_id,
        user_id=user_id,
        data={
            'project_id': project_id,
            'project_deleted': project_deleted
        }
    )


def import_from_eln_file(user_id: int, object_id: int) -> None:
    _store_new_log_entry(
        type=ObjectLogEntryType.IMPORT_FROM_ELN_FILE,
        object_id=object_id,
        user_id=user_id,
        data={}
    )



# coding: utf-8
"""
Create default action types.
"""

import typing

import flask_sqlalchemy

from .utils import table_has_column
from ..actions import ActionType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    pre_translation = table_has_column('action_types', 'name')

    existing_action_type_ids = [
        action_type[0]
        for action_type in db.session.execute(db.text("""
                    SELECT id
                    FROM action_types;
                """)).fetchall()
    ]

    # checks if translatable columns still exist in action_types
    if pre_translation:
        default_action_types = [
            {
                'id': ActionType.SAMPLE_CREATION,
                'name': 'Sample Creation',
                'description': 'These Actions represent processes which create a sample.',
                'object_name': 'Sample',
                'object_name_plural': 'Samples',
                'view_text': 'View Samples',
                'perform_text': 'Create Sample',
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': True,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True
            },
            {
                'id': ActionType.MEASUREMENT,
                'name': 'Measurement',
                'description': 'These Actions represent processes which perform a measurement.',
                'object_name': 'Measurement',
                'object_name_plural': 'Measurements',
                'view_text': 'View Measurements',
                'perform_text': 'Perform Measurement',
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': False,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True
            },
            {
                'id': ActionType.SIMULATION,
                'name': 'Simulation',
                'description': 'These Actions represent processes which run a simulation.',
                'object_name': 'Simulation',
                'object_name_plural': 'Simulations',
                'view_text': 'View Simulations',
                'perform_text': 'Run Simulation',
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': False,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True
            }
        ]

        performed_migration = False
        for action_type in default_action_types:
            # Skip migration by condition
            if action_type['id'] in existing_action_type_ids:
                continue

            # Perform migration
            db.session.execute(db.text("""
                INSERT INTO action_types (id, name, description, object_name, object_name_plural, view_text, perform_text, admin_only, show_on_frontpage, show_in_navbar, enable_labels, enable_files, enable_locations, enable_publications, enable_comments, enable_activity_log, enable_related_objects)
                VALUES (:id, :name, :description, :object_name, :object_name_plural, :view_text, :perform_text, :admin_only, :show_on_frontpage, :show_in_navbar, :enable_labels, :enable_files, :enable_locations, :enable_publications, :enable_comments, :enable_activity_log, :enable_related_objects)
            """), params=action_type)
            performed_migration = True

        return performed_migration
    else:
        default_action_types = [
            {
                'id': ActionType.SAMPLE_CREATION,
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': True,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True,
                'enable_project_link': False
            },
            {
                'id': ActionType.MEASUREMENT,
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': False,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True,
                'enable_project_link': False
            },
            {
                'id': ActionType.SIMULATION,
                'admin_only': False,
                'show_on_frontpage': True,
                'show_in_navbar': True,
                'enable_labels': False,
                'enable_files': True,
                'enable_locations': True,
                'enable_publications': True,
                'enable_comments': True,
                'enable_activity_log': True,
                'enable_related_objects': True,
                'enable_project_link': False
            }
        ]

        scicat_export_types = {
            ActionType.SAMPLE_CREATION: 'SAMPLE',
            ActionType.MEASUREMENT: 'RAW_DATASET',
            ActionType.SIMULATION: 'RAW_DATASET'
        }

        performed_migration = False

        existing_action_type_ids = [
            action_type[0]
            for action_type in db.session.execute(db.text("""
                             SELECT id
                             FROM action_types
                         """)).fetchall()
        ]

        for action_type in default_action_types:
            action_type_id: int = typing.cast(int, action_type['id'])

            # Skip migration by condition
            if action_type_id in existing_action_type_ids:
                continue

            # Perform migration
            db.session.execute(db.text("""
                          INSERT INTO action_types (id, admin_only, show_on_frontpage, show_in_navbar, enable_labels, enable_files, enable_locations, enable_publications, enable_comments, enable_activity_log, enable_related_objects)
                          VALUES (:id, :admin_only, :show_on_frontpage, :show_in_navbar, :enable_labels, :enable_files, :enable_locations, :enable_publications, :enable_comments, :enable_activity_log, :enable_related_objects)
                      """), params=action_type)
            performed_migration = True

            if action_type_id in scicat_export_types:
                db.session.execute(db.text("""
                    UPDATE action_types
                    SET scicat_export_type = :export_type
                    WHERE id = :id
                """), {
                    "id": action_type_id,
                    "export_type": scicat_export_types[action_type_id]
                })

        return performed_migration



# coding: utf-8
"""

"""

from datetime import datetime, timezone
import enum
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, Query, Mapped

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


@enum.unique
class AuthenticationType(enum.Enum):
    LDAP = 1
    EMAIL = 2
    OTHER = 3
    API_TOKEN = 4
    API_ACCESS_TOKEN = 5
    FIDO2_PASSKEY = 6
    FEDERATED_LOGIN = 7
    OIDC = 8


class Authentication(Model):
    __tablename__ = "authentications"

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    login: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)
    type: Mapped[AuthenticationType] = db.Column(db.Enum(AuthenticationType), nullable=False)
    confirmed: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    user: Mapped['User'] = relationship('User', back_populates="authentication_methods")

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Authentication"]]

    def __init__(
            self,
            login: typing.Dict[str, typing.Any],
            authentication_type: AuthenticationType,
            confirmed: bool,
            user_id: int
    ) -> None:
        super().__init__(
            user_id=user_id,
            login=login,
            type=authentication_type,
            confirmed=confirmed
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id})>'


class TwoFactorAuthenticationMethod(Model):
    __tablename__ = "two_factor_authentication_methods"

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    active: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["TwoFactorAuthenticationMethod"]]


class SAMLArtifacts(Model):
    __tablename__ = 'saml_artifacts'

    artifact: Mapped[str] = db.Column(db.String, nullable=False, primary_key=True)
    message: Mapped[str] = db.Column(db.Text, nullable=False)
    since: Mapped[datetime] = db.Column(db.DateTime, nullable=False, default=datetime.now(timezone.utc))

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["SAMLArtifacts"]]

    def __init__(self, artifact: str, message: str):
        super().__init__(artifact=artifact, message=message, since=datetime.now(timezone.utc))

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(artifact={self.artifact})>"


@enum.unique
class SAMLMetadataType(enum.Enum):
    SERVICE_PROVIDER_METADATA = enum.auto()
    IDENTITY_PROVIDER_METADATA = enum.auto()


class SAMLMetadata(Model):
    __tablename__ = 'saml_metadata'

    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey("components.id"), primary_key=True)
    type: Mapped[SAMLMetadataType] = db.Column(db.Enum(SAMLMetadataType), nullable=False, primary_key=True)
    metadata_xml: Mapped[str] = db.Column(db.Text, nullable=False)
    since: Mapped[datetime] = db.Column(db.DateTime, nullable=False, default=datetime.now(timezone.utc))

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["SAMLMetadata"]]

    def __init__(self, component_id: int, metadata_xml: str, type: SAMLMetadataType):
        super().__init__(component_id=component_id, metadata_xml=metadata_xml, since=datetime.now(timezone.utc), type=type)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(component_id={self.component_id}, since={self.since})>"



# coding: utf-8
"""
Replace string description column with JSON column in object_location_assignments table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_info = db.session.execute(db.text("""
        SELECT column_name, data_type
        FROM information_schema.columns
        WHERE table_name = 'object_location_assignments'
    """)).fetchall()
    if ('description', 'json') in column_info:
        return False

    # Perform migration
    existing_data = [
        location_assignment_data
        for location_assignment_data in db.session.execute(db.text("""
            SELECT id, description
            FROM object_location_assignments
        """)).fetchall()
    ]

    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        DROP COLUMN description;
    """))
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ADD description JSON NOT NULL DEFAULT '{}'::json;
    """))
    for id, description in existing_data:
        location_assignment_data = {
            'id': id,
            'description': description
        }
        db.session.execute(db.text("""
            UPDATE object_location_assignments
            SET description = json_build_object('en', :description)
            WHERE id = :id;
        """), params=location_assignment_data)
    return True



# coding: utf-8
"""
Logic module for versioned generic object storage

At its core, this module provides a procedural wrapper of models.objects and
models.versioned_json_object_tables. At the same time, the functions contain
additional logic, e.g. for initial default permissions or object logs.

In practice, the flask-sqlalchemy database engine instance is used. So even
though the underlying models only rely on SQLAlchemy Core and not the ORM,
the functions in this module should be called from within a Flask app context.
"""


import typing
import datetime

import flask
import sqlalchemy

from .components import get_component_by_uuid
from ..models import Objects, Object, Action, ActionType, Permissions
from . import object_log, user_log, object_permissions, errors, users, actions, tags
from .notifications import create_notification_for_being_referenced_by_object_metadata
from .errors import CreatingObjectsDisabledError
from .utils import cache
from ..logic.schemas.utils import data_iter
from .. import db


def create_object(
        action_id: int,
        data: typing.Dict[str, typing.Any],
        user_id: int,
        previous_object_id: typing.Optional[int] = None,
        schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        copy_permissions_object_id: typing.Optional[int] = None,
        permissions_for_group_id: typing.Optional[int] = None,
        permissions_for_project_id: typing.Optional[int] = None,
        permissions_for_all_users: typing.Optional[Permissions] = None,
        validate_data: bool = True,
        data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> Object:
    """
    Creates an object using the given action and its schema. This function
    also handles logging, object references and default object permissions.

    :param action_id: the ID of an existing action
    :param data: the object's data, which must fit to the action's schema
    :param user_id: the ID of the user who created the object
    :param previous_object_id: the ID of the base object
    :param schema: the object schema used for validation. If schema is None
        the action schema is used
    :param copy_permissions_object_id: the ID of an existing object to copy
        the permissions from or None
    :param permissions_for_group_id: the ID of an existing group to give
        permissions to
    :param permissions_for_project_id: the ID of an existing project to give
        permissions to
    :param permissions_for_all_users: permissions to be granted to all
        signed-in users, or None
    :param validate_data: whether the data should be validated
    :param data_validator_arguments: additional keyword arguments to the data
        validator
    :return: the created object
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.CreatingObjectsDisabledError: when creating objects is
        disabled for the given action
    """
    action = actions.get_action(action_id)
    if action.type is not None and action.type.disable_create_objects:
        raise CreatingObjectsDisabledError()
    users.check_user_exists(user_id)
    if data_validator_arguments is None:
        data_validator_arguments = {}
    # unless file names are explicitly provided, assume no files exist yet
    if 'file_names_by_id' not in data_validator_arguments:
        data_validator_arguments['file_names_by_id'] = {}
    object = Objects.create_object(
        data=data,
        schema=schema,
        user_id=user_id,
        action_id=action_id,
        validate_data=validate_data,
        data_validator_arguments=data_validator_arguments
    )
    if copy_permissions_object_id is not None:
        object_permissions.copy_permissions(object.id, copy_permissions_object_id)
        object_permissions.set_user_object_permissions(object.id, user_id, Permissions.GRANT)
    elif permissions_for_group_id is not None:
        object_permissions.set_group_object_permissions(object.id, permissions_for_group_id, Permissions.GRANT)
    elif permissions_for_project_id is not None:
        object_permissions.set_project_object_permissions(object.id, permissions_for_project_id, Permissions.GRANT)
    else:
        object_permissions.set_initial_permissions(object)
    if permissions_for_all_users is not None:
        object_permissions.set_object_permissions_for_all_users(object.id, permissions_for_all_users)
    object_log.create_object(object_id=object.object_id, user_id=user_id, previous_object_id=previous_object_id)
    user_log.create_object(object_id=object.object_id, user_id=user_id)
    _update_object_references(object, user_id=user_id)
    _send_user_references_notifications(object, user_id)
    tags.update_object_tag_usage(object)
    return object


def create_eln_import_object(
        eln_import_id: int,
        eln_object_id: str,
        action_id: typing.Optional[int],
        data: typing.Optional[typing.Dict[str, typing.Any]],
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        importing_user_id: int,
        user_id: typing.Optional[int],
        utc_datetime: typing.Optional[datetime.datetime],
        data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> Object:
    users.check_user_exists(importing_user_id)
    if user_id:
        users.check_user_exists(user_id)
    if data_validator_arguments is None:
        data_validator_arguments = {}
    object = Objects.create_object(
        data=data,
        schema=schema,
        user_id=user_id,
        action_id=action_id,
        utc_datetime=utc_datetime,
        eln_import_id=eln_import_id,
        eln_object_id=eln_object_id,
        data_validator_arguments=data_validator_arguments
    )
    if user_id:
        _update_object_references(object, user_id=user_id)
    object_permissions.set_initial_permissions(object, user_id=importing_user_id)
    object_log.import_from_eln_file(object_id=object.object_id, user_id=importing_user_id)
    user_log.import_from_eln_file(object_id=object.object_id, user_id=importing_user_id)
    tags.update_object_tag_usage(object)
    return object


def insert_fed_object_version(
        fed_object_id: int,
        fed_version_id: int,
        component_id: int,
        action_id: typing.Optional[int],
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        data: typing.Optional[typing.Dict[str, typing.Any]],
        user_id: typing.Optional[int],
        utc_datetime: typing.Optional[datetime.datetime],
        allow_disabled_languages: bool = False,
        get_missing_schema_from_action: bool = True
) -> typing.Optional[Object]:
    """
    Inserts an imported object version into local object versions.

    :param fed_object_id: the federated object ID
    :param fed_version_id: the federated version ID
    :param component_id: the ID of the component
    :param action_id: the ID of an existing action, optional
    :param schema: the object schema used for validation. If schema is None
        the action schema is used, optional
    :param data: the object's data, which must fit to the action's schema, optional
    :param user_id: the ID of the user who created the object, optional
    :param utc_datetime: the creation datetime of the version to insert
    :param allow_disabled_languages: whether disabled languages may be allowed
        in data
    :param get_missing_schema_from_action: whether to use an action schema (if available) when None is passed for schema
    :return: the created object
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    if action_id is not None:
        actions.check_action_exists(action_id)
    if user_id is not None:
        users.check_user_exists(user_id)
    object = Objects.insert_fed_object_version(
        data=data,
        schema=schema,
        user_id=user_id,
        action_id=action_id,
        utc_datetime=utc_datetime,
        fed_object_id=fed_object_id,
        fed_version_id=fed_version_id,
        component_id=component_id,
        allow_disabled_languages=allow_disabled_languages,
        get_missing_schema_from_action=get_missing_schema_from_action
    )
    if object is not None:
        tags.update_object_tag_usage(object)
    return object


def create_object_batch(
        action_id: int,
        data_sequence: typing.Sequence[typing.Dict[str, typing.Any]],
        user_id: int,
        copy_permissions_object_id: typing.Optional[int] = None,
        permissions_for_group_id: typing.Optional[int] = None,
        permissions_for_project_id: typing.Optional[int] = None,
        permissions_for_all_users: typing.Optional[Permissions] = None,
        validate_data: bool = True,
        data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> typing.Sequence[Object]:
    """
    Creates a batch of objects using the given action and its schema. This
    function also handles logging, object references and default object
    permissions. When creating multiple objects for the same action and user
    this function should be used instead of repeatedly calling create_object.

    :param action_id: the ID of an existing action
    :param data_sequence: a sequence containing the objects' data, which must
        fit to the action's schema
    :param user_id: the ID of the user who created the objects
    :param copy_permissions_object_id: the ID of an existing object to copy
        the permissions from or None
    :param permissions_for_group_id: the ID of an existing group to give
        permissions to
    :param permissions_for_project_id: the ID of an existing project to give
        permissions to
    :param permissions_for_all_users: permissions to be granted to all
        signed-in users, or None
    :param validate_data: whether the data should be validated
    :param data_validator_arguments: additional keyword arguments to the data
        validator
    :return: the created objects
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    objects: typing.List[Object] = []
    actions.check_action_exists(action_id)
    users.check_user_exists(user_id)
    if data_validator_arguments is None:
        data_validator_arguments = {}
    # unless file names are explicitly provided, assume no files exist yet
    if 'file_names_by_id' not in data_validator_arguments:
        data_validator_arguments['file_names_by_id'] = {}
    try:
        for data in data_sequence:
            object = Objects.create_object(
                data=data,
                schema=None,
                user_id=user_id,
                action_id=action_id,
                validate_data=validate_data,
                data_validator_arguments=data_validator_arguments
            )
            objects.append(object)
    finally:
        if objects:
            # objects created before the integrity error must still be accessible
            batch_object_ids = [object.id for object in objects]
            user_log.create_batch(user_id=user_id, batch_object_ids=batch_object_ids)
            for object in objects:
                _update_object_references(object, user_id=user_id)
                _send_user_references_notifications(object, user_id)
                if copy_permissions_object_id is not None:
                    object_permissions.copy_permissions(object.id, copy_permissions_object_id)
                    object_permissions.set_user_object_permissions(object.id, user_id, Permissions.GRANT)
                elif permissions_for_group_id is not None:
                    object_permissions.set_group_object_permissions(object.id, permissions_for_group_id, Permissions.GRANT)
                elif permissions_for_project_id is not None:
                    object_permissions.set_project_object_permissions(object.id, permissions_for_project_id, Permissions.GRANT)
                else:
                    object_permissions.set_initial_permissions(object)
                if permissions_for_all_users is not None:
                    object_permissions.set_object_permissions_for_all_users(object.id, permissions_for_all_users)
                object_log.create_batch(object_id=object.object_id, user_id=user_id, batch_object_ids=batch_object_ids)
                tags.update_object_tag_usage(object)
    return objects


def update_object(
        object_id: int,
        data: typing.Optional[typing.Dict[str, typing.Any]],
        user_id: int,
        schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None,
        create_log_entries: bool = True,
        utc_datetime: typing.Optional[datetime.datetime] = None
) -> None:
    """
    Updates the object to a new version. This function also handles logging
    and object references.

    :param object_id: the ID of the existing object
    :param data: the object's new data, which must fit to the object's schema
    :param user_id: the ID of the user who updated the object
    :param schema: the schema for the new object data
    :param data_validator_arguments: additional keyword arguments to the data
        validator
    :param create_log_entries: whether log entries should be created
    :param utc_datetime: the datetime of the update, or None
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    if data_validator_arguments is None:
        data_validator_arguments = {}
    if 'file_names_by_id' not in data_validator_arguments:
        # local import to avoid circular imports
        from .files import get_file_names_by_id_for_object
        data_validator_arguments['file_names_by_id'] = {
            file_id: file_names[0]
            for file_id, file_names in get_file_names_by_id_for_object(object_id).items()
        }
    object = Objects.update_object(
        object_id=object_id,
        data=data,
        schema=schema,
        user_id=user_id,
        data_validator_arguments=data_validator_arguments,
        utc_datetime=utc_datetime
    )
    if object is None:
        raise errors.ObjectDoesNotExistError()
    if create_log_entries:
        user_log.edit_object(user_id=user_id, object_id=object.object_id, version_id=object.version_id)
        object_log.edit_object(object_id=object.object_id, user_id=user_id, version_id=object.version_id)
    _update_object_references(object, user_id=user_id)
    _send_user_references_notifications(object, user_id)
    tags.update_object_tag_usage(object)


def update_object_version(
        object_id: int,
        version_id: int,
        action_id: typing.Optional[int],
        data: typing.Optional[typing.Dict[str, typing.Any]],
        user_id: typing.Optional[int],
        schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        allow_disabled_languages: bool = False,
        get_missing_schema_from_action: bool = True
) -> Object:
    object = Objects.update_object_version(
        object_id=object_id,
        version_id=version_id,
        action_id=action_id,
        schema=schema,
        data=data,
        user_id=user_id,
        utc_datetime=utc_datetime,
        allow_disabled_languages=allow_disabled_languages,
        get_missing_schema_from_action=get_missing_schema_from_action
    )
    if object is None:
        check_object_version_exists(object_id, version_id)
        raise errors.ObjectNotFederatedError()
    tags.update_object_tag_usage(object, new_subversion=True)

    return object


def restore_object_version(object_id: int, version_id: int, user_id: int) -> None:
    """
    Reverts the changes made to an object and restores a specific version of
    it, while keeping the version history. This function merely adds a new
    version which sets the data and schema to those of the given version.

    :param object_id: the ID of the existing object
    :param version_id: the ID of the object's existing version
    :param user_id: the ID of the user who restored the object version
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    object = Objects.restore_object_version(
        object_id=object_id,
        version_id=version_id,
        user_id=user_id
    )
    if object is None:
        # ensure the object actually exists
        check_object_version_exists(object_id=object_id, version_id=version_id)
        return
    user_log.restore_object_version(user_id=user_id, object_id=object_id, restored_version_id=version_id, version_id=object.version_id)
    object_log.restore_object_version(object_id=object_id, user_id=user_id, restored_version_id=version_id, version_id=object.version_id)
    tags.update_object_tag_usage(object)


@cache
def check_object_exists(object_id: int) -> None:
    """
    Ensures that an object with the specific ID exists.

    :param object_id: the ID of the existing object

    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    if not Objects.is_existing_object(object_id=object_id):
        raise errors.ObjectDoesNotExistError()


@cache
def check_object_version_exists(object_id: int, version_id: int) -> None:
    """
    Ensures that an object version with the given IDs exists.

    :param object_id: the ID of the existing object
    :param version_id: the ID of the existing version for that object

    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when no object version
        with the given object ID and version ID combination exists
    """
    if not Objects.is_existing_object_version(object_id=object_id, version_id=version_id):
        check_object_exists(object_id)
        raise errors.ObjectVersionDoesNotExistError()


def get_object(object_id: int, version_id: typing.Optional[int] = None) -> Object:
    """
    Returns either the current or a specific version of the object.

    :param object_id: the ID of the existing object
    :param version_id: the ID of the object's existing version
    :return: the object with the given object and version IDs
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    """
    if version_id is None:
        object = Objects.get_current_object(
            object_id=object_id
        )
        if object is None:
            raise errors.ObjectDoesNotExistError()
    else:
        object = Objects.get_object_version(
            object_id=object_id,
            version_id=version_id
        )
        if object is None:
            if Objects.get_current_object(object_id=object_id) is None:
                raise errors.ObjectDoesNotExistError()
            else:
                raise errors.ObjectVersionDoesNotExistError()
    return object


def get_fed_object(fed_object_id: int, component_id: int, fed_version_id: typing.Optional[int] = None) -> Object:
    """
    Returns either the current or a specific version of the federated object.

    :param fed_object_id: the federated ID of the existing object
    :param component_id: the ID of the existing component
    :param fed_version_id: the ID of the object's existing version
    :return: the object with the given object and version IDs
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    """
    if fed_version_id is None:
        object = Objects.get_current_fed_object(
            fed_object_id=fed_object_id,
            component_id=component_id
        )
        if object is None:
            raise errors.ObjectDoesNotExistError()
    else:
        object = Objects.get_fed_object_version(
            component_id=component_id,
            fed_object_id=fed_object_id,
            fed_version_id=fed_version_id
        )
        if object is None:
            if Objects.get_current_fed_object(fed_object_id=fed_object_id, component_id=component_id) is None:
                raise errors.ObjectDoesNotExistError()
            else:
                raise errors.ObjectVersionDoesNotExistError()
    return object


def get_object_versions(object_id: int) -> typing.List[Object]:
    """
    Returns all versions of an object, sorted from oldest to newest.

    :param object_id: the ID of the existing object
    :return: the object versions
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    object_versions = Objects.get_object_versions(object_id=object_id)
    if not object_versions:
        raise errors.ObjectDoesNotExistError()
    return object_versions


def get_objects(
        filter_func: typing.Callable[[typing.Any], typing.Any] = lambda data: True,
        action_filter: typing.Optional[sqlalchemy.sql.ColumnElement[bool]] = None,
        **kwargs: typing.Any
) -> typing.List[Object]:
    """
    Returns all objects, optionally after filtering the objects by their data
    or by their actions' information.

    :param filter_func: a lambda that may return an SQLAlchemy filter when
        given the object table's data column
    :param action_filter: a SQLAlchemy comparator, used to query only objects
        created by specific actions
    :param kwargs: additional parameters to be passed to Objects.get_current_objects
    :return: a list of all objects or those matching the given filters
    """
    if action_filter is None:
        action_table = None
    else:
        action_table = Action.__table__
    return Objects.get_current_objects(
        filter_func=filter_func,
        action_table=action_table,
        action_filter=action_filter,
        **kwargs
    )


@typing.overload
def find_object_references(
        *,
        object_id: int,
        version_id: int,
        object_data: typing.Optional[typing.Dict[str, typing.Any]],
        find_previous_referenced_object_ids: bool = True,
        include_fed_references: typing.Literal[True]
) -> typing.Sequence[typing.Tuple[typing.Tuple[int, typing.Optional[str]], typing.Optional[int], str]]:
    ...


@typing.overload
def find_object_references(
        *,
        object_id: int,
        version_id: int,
        object_data: typing.Optional[typing.Dict[str, typing.Any]],
        find_previous_referenced_object_ids: bool = True,
        include_fed_references: typing.Literal[False] = False
) -> typing.Sequence[typing.Tuple[int, typing.Optional[int], str]]:
    ...


def find_object_references(
        *,
        object_id: int,
        version_id: int,
        object_data: typing.Optional[typing.Dict[str, typing.Any]],
        find_previous_referenced_object_ids: bool = True,
        include_fed_references: bool = False,
) -> typing.Sequence[typing.Tuple[typing.Union[int, typing.Tuple[int, typing.Optional[str]]], typing.Optional[int], str]]:
    """
    Searches for references to other objects.

    :param object_id: the object ID
    :param version_id: the object version ID
    :param object_data: the object data
    :param find_previous_referenced_object_ids: whether to find previous referenced object ids
    :param include_fed_references: whether references on other components should be included
    """
    if object_data is None:
        return []
    referenced_object_ids = []
    referenced_object_id: typing.Union[int, typing.Tuple[int, typing.Optional[str]]]
    for path, data in data_iter(data=object_data, filter_property_types={'sample', 'measurement', 'object_reference'}):
        if isinstance(data, dict) and data.get('object_id') is not None:
            if 'component_uuid' in data and data['component_uuid'] != flask.current_app.config['FEDERATION_UUID']:
                try:
                    component = get_component_by_uuid(data['component_uuid'])
                    if include_fed_references:
                        referenced_object_id = (get_fed_object(data['object_id'], component.id).object_id, None)
                    else:
                        referenced_object_id = get_fed_object(data['object_id'], component.id).object_id
                except errors.ComponentDoesNotExistError:
                    if include_fed_references:
                        referenced_object_id = (data['object_id'], data['component_uuid'])
                    else:
                        continue
                except errors.ObjectDoesNotExistError:
                    if include_fed_references:
                        referenced_object_id = (data['object_id'], data['component_uuid'])
                    else:
                        continue
                previous_referenced_object_id = None
            elif 'eln_source_url' in data:
                continue
            else:
                if include_fed_references:
                    referenced_object_id = (data['object_id'], None)
                else:
                    referenced_object_id = data['object_id']
                previous_referenced_object_id = None
                if find_previous_referenced_object_ids and version_id > 0:
                    previous_object_version = get_object(object_id, version_id - 1)
                    previous_data = previous_object_version.data
                    try:
                        for path_element in path:
                            previous_data = previous_data[path_element]  # type: ignore
                    except (KeyError, IndexError):
                        pass
                    else:
                        if previous_data is not None and 'object_id' in previous_data and previous_data['object_id'] is not None:
                            previous_referenced_object_id = previous_data['object_id']
            referenced_object_ids.append((referenced_object_id, previous_referenced_object_id, data['_type']))
    return referenced_object_ids


def _update_object_references(object: Object, user_id: int) -> None:
    """
    Searches for references to other objects and updates these accordingly.

    At this time, only measurements or samples referencing other measurements
    or samples will be handled, adding an entry to the object's log about
    being used in a sample or measurement.

    :param object: the updated (or newly created) object
    :param user_id: the user who caused the object update or creation
    """
    if object.action_id is None:
        action_type_id = None
    else:
        action_type_id = actions.get_action(object.action_id).type_id
    for referenced_object_id, previous_referenced_object_id, schema_type in find_object_references(object_id=object.object_id, version_id=object.version_id, object_data=object.data):
        if referenced_object_id != previous_referenced_object_id:
            if action_type_id == ActionType.MEASUREMENT and schema_type == 'sample':
                object_log.use_object_in_measurement(object_id=referenced_object_id, user_id=user_id, measurement_id=object.object_id)
            elif action_type_id == ActionType.SAMPLE_CREATION and schema_type == 'sample':
                object_log.use_object_in_sample(object_id=referenced_object_id, user_id=user_id, sample_id=object.object_id)
            else:
                object_log.reference_object_in_metadata(object_id=referenced_object_id, user_id=user_id, referencing_object_id=object.object_id)


def _send_user_references_notifications(object: Object, user_id: int) -> None:
    """
    Searches for references to users and notifies them accordingly.

    :param object: the updated (or newly created) object
    :param user_id: the user who caused the object update or creation
    """

    for referenced_user_id, previous_referenced_user_id in find_user_references(object):
        if referenced_user_id not in {user_id, previous_referenced_user_id}:
            create_notification_for_being_referenced_by_object_metadata(referenced_user_id, object.object_id)


def find_user_references(object: Object, find_previous_referenced_user_ids: bool = True) -> typing.List[typing.Tuple[int, typing.Optional[int]]]:
    """
    Searches for references to users.

    :param object: the updated (or newly created) object
    :param find_previous_referenced_user_ids: whether to find previous
        referenced user ids
    """
    if object.data is None:
        return []
    referenced_user_ids = []
    for path, data in data_iter(data=object.data, filter_property_types={'user'}):
        if isinstance(data, dict) and data.get('user_id') is not None:
            referenced_user_id = data['user_id']
            previous_referenced_user_id = None
            if find_previous_referenced_user_ids and object.version_id > 0:
                previous_object_version = get_object(object.object_id, object.version_id - 1)
                previous_data = previous_object_version.data
                try:
                    for path_element in path:
                        previous_data = previous_data[path_element]  # type: ignore
                except (KeyError, IndexError):
                    pass
                else:
                    if previous_data is not None and 'user_id' in previous_data and previous_data['user_id'] is not None:
                        previous_referenced_user_id = previous_data['user_id']
            referenced_user_ids.append((referenced_user_id, previous_referenced_user_id))
    return referenced_user_ids


def get_current_object_version_id(object_id: int) -> int:
    """
    Get the ID of the current version of an object.

    :param object_id: the ID of an existing object
    :return: the object's current version ID
    :raises errors.ObjectDoesNotExistError: if the object does not exist
    """
    version_id = Objects.get_current_object_version_id(object_id)
    if version_id is None:
        raise errors.ObjectDoesNotExistError()
    return version_id


def get_action_ids_for_object_ids(
        object_ids: typing.Sequence[int]
) -> typing.Dict[int, typing.Optional[int]]:
    """
    Get the action IDs for a list of object IDs.

    :param object_ids: the IDs of existing objects
    :return: the objects' action IDs
    """
    return Objects.get_action_ids_for_object_ids(object_ids)


def get_max_object_id() -> typing.Optional[int]:
    """
    Get the maximum object ID.

    :return: the largest existing object ID, or None if no objects exist
    """
    return typing.cast(typing.Optional[int], db.session.query(db.func.max(Objects.object_id_column)).scalar())



# coding: utf-8
"""
Add missing NOT NULL constraints to type column in notification_mode_for_types table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('notification_mode_for_types', 'type'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        DELETE FROM notification_mode_for_types
        WHERE type IS NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE notification_mode_for_types
        ALTER COLUMN type SET NOT NULL
    """))
    return True



# coding: utf-8
"""

"""
import datetime
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User
    from .group_categories import GroupCategory


association_table = db.Table(
    'user_group_memberships',
    db.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id'))
)


class Group(Model):
    __tablename__ = 'groups'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    description: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    members: Mapped[typing.List['User']] = relationship("User", secondary=association_table, back_populates="groups")
    categories: Mapped[typing.List['GroupCategory']] = relationship('GroupCategory', secondary='basic_group_category_association_table', back_populates='basic_groups')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Group"]]


class GroupInvitation(Model):
    __tablename__ = 'group_invitations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    inviter_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    accepted: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    revoked: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())  # TODO: migration

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupInvitation"]]



# coding: utf-8
"""
RESTful API for SampleDB
"""

import base64
import typing

import flask

from .authentication import object_permissions_required
from ..utils import Resource, ResponseData
from ...logic import errors
from ...logic.actions import get_action
from ...logic.files import create_database_file, create_local_file_reference, create_url_file, File, get_file_for_object, get_files_for_object, SUPPORTED_HASH_ALGORITHMS, DEFAULT_HASH_ALGORITHM
from ...logic.objects import get_object
from ...logic.utils import parse_url
from ...models import Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def file_info_to_json(file_info: File, include_content: bool = True) -> typing.Dict[str, typing.Any]:
    file_json = {
        'object_id': file_info.object_id,
        'file_id': file_info.id,
        'storage': file_info.storage
    }
    if file_info.storage == 'database':
        file_json.update({
            'original_file_name': file_info.original_file_name
        })
        if include_content:
            with file_info.open() as f:
                base64_content = base64.b64encode(f.read())
            file_json.update({
                'base64_content': base64_content.decode('utf-8')
            })
        if file_info.data:
            file_json['hash'] = file_info.data.get('hash')
        if file_info.preview_image_binary_data and file_info.preview_image_mime_type:
            file_json['base64_preview_image'] = base64.b64encode(file_info.preview_image_binary_data).decode('utf-8')
            file_json['preview_image_mime_type'] = file_info.preview_image_mime_type
    if file_info.storage == 'url' and file_info.url is not None:
        file_json.update({
            'url': file_info.url
        })
    return file_json


class ObjectFile(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, file_id: int) -> ResponseData:
        try:
            file_info = get_file_for_object(object_id=object_id, file_id=file_id)
            if file_info is None:
                raise errors.FileDoesNotExistError()
        except errors.FileDoesNotExistError:
            return {
                "message": f"file {file_id} of object {object_id} does not exist"
            }, 404
        if file_info.is_hidden:
            return {
                "message": f"file {file_id} of object {object_id} has been hidden"
            }, 403
        return file_info_to_json(file_info)


class ObjectFiles(Resource):
    @object_permissions_required(Permissions.WRITE)
    def post(self, object_id: int) -> ResponseData:
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        object = get_object(object_id=object_id)
        if object.action_id is not None:
            action = get_action(action_id=object.action_id)
            if action.type is None or not action.type.enable_files:
                return {
                    "message": "Adding files is not enabled for objects of this type"
                }, 403
        if 'object_id' in request_json:
            if request_json['object_id'] != object.object_id:
                return {
                    "message": f"object_id must be {object.object_id}"
                }, 400
        if 'storage' not in request_json:
            return {
                "message": "storage must be set"
            }, 400
        if 'hash' in request_json:
            if not isinstance(request_json['hash'], dict):
                return {
                    "message": "hash must be a JSON object"
                }, 400
            if set(request_json['hash'].keys()) != {'hexdigest', 'algorithm'} or not isinstance(request_json['hash']['algorithm'], str) or not isinstance(request_json['hash']['hexdigest'], str):
                return {
                    "message": "hash must contain algorithm and hexdigest as strings"
                }, 400
            hash_algorithm = request_json['hash']['algorithm']
            if hash_algorithm not in SUPPORTED_HASH_ALGORITHMS:
                return {
                    "message": "only the following hash algorithms are supported: " + ", ".join(
                        SUPPORTED_HASH_ALGORITHMS)
                }, 400
            if any(c not in '0123456789abcdef' for c in request_json['hash']['hexdigest']):
                return {
                    "message": "hash hexdigest be a lowercase string of hex characters"
                }, 400
        storage = request_json['storage']
        if storage == 'database':
            for key in request_json:
                if key not in {'object_id', 'storage', 'original_file_name', 'base64_content', 'hash', 'base64_preview_image', 'preview_image_mime_type'}:
                    return {
                        "message": f"invalid key '{key}'"
                    }, 400
            if 'original_file_name' not in request_json or not request_json['original_file_name']:
                return {
                    "message": "original_file_name must be set for files with database storage"
                }, 400
            original_file_name = request_json['original_file_name']
            if 'base64_content' not in request_json:
                return {
                    "message": "base64_content must be set for files with database storage"
                }, 400
            base64_content = request_json['base64_content']
            try:
                content = base64.b64decode(base64_content.encode('utf-8'), validate=True)
            except Exception:
                return {
                    "message": "base64_content must be base64 encoded"
                }, 400
            if 'hash' in request_json:
                hash = File.HashInfo.from_binary_data(
                    algorithm=request_json['hash']['algorithm'],
                    binary_data=content
                )
                if hash is None:
                    return {
                        "message": "invalid hash data"
                    }, 400
                if request_json['hash']['hexdigest'] != hash.hexdigest:
                    return {
                        "message": "hash hexdigest mismatch, expected: " + hash.hexdigest
                    }, 400
            else:
                hash = File.HashInfo.from_binary_data(
                    algorithm=DEFAULT_HASH_ALGORITHM,
                    binary_data=content
                )
            if 'base64_preview_image' in request_json:
                base64_preview_image = request_json['base64_preview_image']
                try:
                    preview_image_binary_data = base64.b64decode(base64_preview_image.encode('utf-8'), validate=True)
                except Exception:
                    return {
                        "message": "base64_preview_image must be base64 encoded"
                    }, 400
                preview_image_mime_type = request_json.get('preview_image_mime_type', '').lower()
                supported_mime_types = sorted(flask.current_app.config['MIME_TYPES'].values())
                if preview_image_mime_type not in supported_mime_types:
                    return {
                        "message": "preview_image_mime_type must be one of: " + ", ".join(supported_mime_types)
                    }, 400
            else:
                preview_image_binary_data = None
                preview_image_mime_type = None
            file = create_database_file(
                object_id=object_id,
                user_id=flask.g.user.id,
                file_name=original_file_name,
                save_content=lambda stream: typing.cast(None, stream.write(content)),
                hash=hash,
                preview_image_binary_data=preview_image_binary_data,
                preview_image_mime_type=preview_image_mime_type
            )
        elif storage == 'local_reference':
            for key in request_json:
                if key not in {'object_id', 'storage', 'filepath', 'hash'}:
                    return {
                        "message": f"invalid key '{key}'"
                    }, 400
            if 'filepath' not in request_json or not request_json['filepath']:
                return {
                    "message": "filepath must be set for files with local_reference storage"
                }, 400
            try:
                file = create_local_file_reference(
                    object_id=object_id,
                    user_id=flask.g.user.id,
                    filepath=request_json['filepath'],
                    hash=File.HashInfo(
                        algorithm=request_json['hash']['algorithm'],
                        hexdigest=request_json['hash']['hexdigest']
                    ) if 'hash' in request_json else None
                )
            except errors.UnauthorizedRequestError:
                return {
                    "message": "user not authorized to add this path"
                }, 403
        elif storage == 'url':
            for key in request_json:
                if key not in {'object_id', 'storage', 'url'}:
                    return {
                        "message": f"invalid key '{key}'"
                    }, 400
            if 'url' not in request_json:
                return {
                    "message": "url must be set for files with url storage"
                }, 400
            url = request_json['url']
            try:
                parse_url(url)
            except errors.InvalidURLError:
                return {
                    "message": "url must be a valid url"
                }, 400
            except errors.URLTooLongError:
                return {
                    "message": "url exceeds length limit"
                }, 400
            except errors.InvalidIPAddressError:
                return {
                    "message": "url contains an invalid IP address"
                }, 400
            except errors.InvalidPortNumberError:
                return {
                    "message": "url contains an invalid port number"
                }, 400
            file = create_url_file(
                object_id=object_id,
                user_id=flask.g.user.id,
                url=url
            )
        else:
            return {
                "message": "storage must be 'local_reference', 'database' or 'url'"
            }, 400
        file_url = flask.url_for(
            'api.object_file',
            object_id=file.object_id,
            file_id=file.id,
            _external=True
        )
        return flask.redirect(file_url, code=201)

    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int) -> ResponseData:
        return [
            file_info_to_json(file_info, include_content=False)
            for file_info in get_files_for_object(object_id)
            if not file_info.is_hidden
        ]



# coding: utf-8
"""

"""

import importlib.util
import os
import string
import typing
import types


def _find_script_modules() -> typing.Dict[str, types.ModuleType]:
    """
    Find, import and return all .py files located in the scripts directory.
    """
    scripts_path = os.path.abspath(os.path.dirname(__file__))
    script_modules = {}
    for file_name in os.listdir(scripts_path):
        module_name, file_extension = os.path.splitext(file_name)
        # only import Python files
        if file_extension != '.py':
            continue
        # prevent importing files with invalid module names
        if not all(c in string.ascii_letters + string.digits + '_' for c in module_name) or module_name.startswith('__'):
            continue
        module_path = os.path.join(scripts_path, file_name)
        spec = importlib.util.spec_from_file_location(f'sampledb.scripts.{module_name}', module_path)
        if not spec or not spec.loader:
            continue
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        script_modules[module_name] = module
    return script_modules


script_modules = _find_script_modules()
globals().update(script_modules)

script_documentation = {
    script_name: script_module.__doc__
    for script_name, script_module
    in script_modules.items()
}

script_functions = {
    script_name: script_module.main
    for script_name, script_module
    in script_modules.items()
}



# coding: utf-8
"""

"""

import typing
import datetime

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .components import Component
from .objects import Objects
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


class ObjectShare(Model):
    __tablename__ = 'object_shares'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False, primary_key=True)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Component.id), nullable=False, primary_key=True)
    policy: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    component: Mapped[Component] = relationship('Component')
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
    user: Mapped[typing.Optional['User']] = relationship('User')
    import_status: Mapped[typing.Optional[typing.Dict[str, typing.Any]]] = db.Column(postgresql.JSONB)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ObjectShare"]]

    def __init__(
            self,
            object_id: int,
            component_id: int,
            policy: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            user_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            object_id=object_id,
            component_id=component_id,
            policy=policy,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            user_id=user_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(object_id={self.object_id}, component_id={self.component_id}, policy={self.policy}, utc_datetime={self.utc_datetime})>'



# coding: utf-8
"""
Add the scicat_export_type column to the action_types table.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..actions import ActionType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('action_types', 'scicat_export_type'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE action_types
        ADD scicat_export_type scicatexporttype
    """))

    scicat_export_types = {
        ActionType.SAMPLE_CREATION: 'SAMPLE',
        ActionType.MEASUREMENT: 'RAW_DATASET',
        ActionType.SIMULATION: 'RAW_DATASET'
    }
    for action_type_id, export_type in scicat_export_types.items():
        db.session.execute(db.text("""
            UPDATE action_types
            SET scicat_export_type = :export_type
            WHERE id = :id
        """), {
            "id": action_type_id,
            "export_type": export_type
        })
    return True



# coding: utf-8
"""
Splits actions into action and action_translations
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..languages import Language


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not table_has_column('actions', 'name'):
        return False

    # Perform migration
    existing_data = [
        action_data
        for action_data in db.session.execute(db.text("""
            SELECT id, name, description, short_description
            FROM actions
        """)).fetchall()
    ]

    db.session.execute(db.text("""
            ALTER TABLE actions
            DROP COLUMN name,
            DROP COLUMN description,
            DROP COLUMN short_description
        """))

    for action_id, name, description, short_description in existing_data:
        # Perform migration
        translation = {
            'action_id': action_id,
            'language_id': Language.ENGLISH,
            'name': name,
            'description': description,
            'short_description': short_description,
        }
        db.session.execute(db.text("""
                      INSERT INTO action_translations (action_id, language_id, name, description, short_description)
                      VALUES (:action_id, :language_id, :name, :description, :short_description)
                  """), params=translation)

    return True



# coding: utf-8
"""
Change datetime_format_moment_output column data type to varchar in languages table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE languages
        ALTER COLUMN datetime_format_moment_output SET DATA TYPE VARCHAR
    """))
    db.session.execute(db.text("""
        ALTER TABLE languages
        ALTER COLUMN datetime_format_moment_output SET DEFAULT 'lll'::VARCHAR
    """))
    return True



# coding: utf-8
"""
Add entries to the public_actions table for all actions without a user_id.

Previously, these actions were implicitly public.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    common_actions_result = db.session.execute(db.text("""
        SELECT id
        FROM actions
        WHERE user_id IS NULL
    """))
    common_actions = {row[0] for row in common_actions_result}

    table_exists = db.session.execute(db.text("""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = 'public_actions'
    """)).fetchall()
    if table_exists:
        public_actions_result = db.session.execute(db.text("""
            SELECT action_id
            FROM public_actions
        """))
        public_actions = {row[0] for row in public_actions_result}
        for action_id in common_actions - public_actions:
            db.session.execute(db.text("""
                INSERT INTO public_actions
                (action_id)
                VALUES
                (:action_id)
            """), {'action_id': action_id})
    else:
        all_user_action_permissions_result = db.session.execute(db.text("""
            SELECT action_id, permissions
            FROM all_user_action_permissions
        """)).fetchall()
        all_user_action_permissions = {
            action_id: permissions
            for action_id, permissions in all_user_action_permissions_result
        }
        for action_id in common_actions:
            if action_id not in all_user_action_permissions:
                db.session.execute(db.text("""
                    INSERT INTO public_actions
                    (action_id, permissions)
                    VALUES (:action_id, 'READ')
                """), {'action_id': action_id})
    return True
